@@ -73,6 +73,11 @@
 #include <termios.h>
 
 #include "config.h"
+#include "extern.h"
+
+#if defined (HAVE_SYS_WAIT_H)
+#  include <sys/wait.h>
+#endif
 
 #ifdef READLINE_LIBRARY
 #  include "readline.h"
@@ -102,16 +107,14 @@
 #  else
 #    define memmove(d, s, n)	memcpy(d, s, n)
 #  endif
-#else
-#  define memmove(d, s, n)	memcpy(d, s, n)
 #endif
 
 #define APPLICATION_NAME "rlfe"
 
 static int in_from_inferior_fd;
 static int out_to_inferior_fd;
-static void set_edit_mode ();
-static void usage_exit ();
+static void set_edit_mode (void);
+static void usage_exit (void);
 static char *hist_file = 0;
 static int  hist_size = 0;
 
@@ -262,7 +265,7 @@ static void maybe_emphasize_input (int on)
   if (on == current_emphasize_input
       || (on && ! do_emphasize_input))
     return;
-  fprintf (rl_outstream, on ? start_input_mode : end_input_mode);
+  fprintf (rl_outstream, "%s", on ? start_input_mode : end_input_mode);
   fflush (rl_outstream);
   current_emphasize_input = on;
 }
@@ -273,7 +276,7 @@ null_prep_terminal (int meta)
 }
 
 static void
-null_deprep_terminal ()
+null_deprep_terminal (void)
 {
   maybe_emphasize_input (0);
 }
@@ -542,7 +545,14 @@ main(int argc, char** argv)
 #endif
 
   t.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOCTL | ECHOE | \
-		 ECHOK | ECHOKE | ECHONL | ECHOPRT );
+		 ECHOK | ECHONL
+#if defined (ECHOKE)
+		| ECHOKE
+#endif
+#if defined (ECHOPRT)
+		| ECHOPRT
+#endif
+		);
   t.c_iflag &= ~ICRNL;
   t.c_iflag |= IGNBRK;
   t.c_cc[VMIN] = 1;
@@ -588,7 +598,7 @@ main(int argc, char** argv)
 	}
       if (FD_ISSET (in_from_tty_fd, &in_set))
 	{
-	  extern int readline_echoing_p;
+	  extern int _rl_echoing_p;
 	  struct termios term_master;
 	  int do_canon = 1;
 	  int do_icrnl = 1;
@@ -605,9 +615,9 @@ main(int argc, char** argv)
 	    {
 	      do_canon = (term_master.c_lflag & ICANON) != 0;
 	      do_icrnl = (term_master.c_lflag & ICRNL) != 0;
-	      readline_echoing_p = (term_master.c_lflag & ECHO) != 0;
+	      _rl_echoing_p = (term_master.c_lflag & ECHO) != 0;
 	      DPRINT1 ("echo,canon,crnl:%03d\n",
-		       100 * readline_echoing_p
+		       100 * _rl_echoing_p
 		       + 10 * do_canon
 		       + 1 * do_icrnl);
 	    }
@@ -744,7 +754,7 @@ main(int argc, char** argv)
     }
 }
 
-static void set_edit_mode ()
+static void set_edit_mode (void)
 {
   int vi = 0;
   char *shellopts;
@@ -757,7 +767,7 @@ static void set_edit_mode ()
 	  vi = 1;
 	  break;
 	}
-      shellopts = index (shellopts + 1, ':');
+      shellopts = strchr (shellopts + 1, ':');
     }
 
   if (!vi)
@@ -773,7 +783,7 @@ static void set_edit_mode ()
 }
 
 
-static void usage_exit ()
+static void usage_exit (void)
 {
   fprintf (stderr, "Usage: rlfe [-h histfile] [-s size] cmd [arg1] [arg2] ...\n\n");
   exit (1);
