@@ -1,6 +1,6 @@
 /* histfile.c - functions to manipulate the history file. */
 
-/* Copyright (C) 1989-2019 Free Software Foundation, Inc.
+/* Copyright (C) 1989-2019,2023-2025 Free Software Foundation, Inc.
 
    This file contains the GNU History Library (History), a set of
    routines for managing the text of previously typed lines.
@@ -114,8 +114,6 @@ extern int errno;
 #  define PATH_MAX	1024	/* default */
 #endif
 
-extern void _hs_append_history_line PARAMS((int, const char *));
-
 /* history file version; currently unused */
 int history_file_version = 1;
 
@@ -141,11 +139,11 @@ int history_lines_written_to_file = 0;
    for more extensive tests. */
 #define HIST_TIMESTAMP_START(s)		(*(s) == history_comment_char && isdigit ((unsigned char)(s)[1]) )
 
-static char *history_backupfile PARAMS((const char *));
-static char *history_tempfile PARAMS((const char *));
-static int histfile_backup PARAMS((const char *, const char *));
-static int histfile_restore PARAMS((const char *, const char *));
-static int history_rename PARAMS((const char *, const char *));
+static char *history_backupfile (const char *);
+static char *history_tempfile (const char *);
+static int histfile_backup (const char *, const char *);
+static int histfile_restore (const char *, const char *);
+static int history_rename (const char *, const char *);
 
 /* Return the string that should be used in the place of this
    filename.  This only matters when you don't specify the
@@ -155,7 +153,7 @@ history_filename (const char *filename)
 {
   char *return_val;
   const char *home;
-  int home_len;
+  size_t home_len;
 
   return_val = filename ? savestring (filename) : (char *)NULL;
 
@@ -170,9 +168,8 @@ history_filename (const char *filename)
 
   if (home == 0)
     return (NULL);
-  else
-    home_len = strlen (home);
 
+  home_len = strlen (home);
   return_val = (char *)xmalloc (2 + home_len + 8); /* strlen(".history") == 8 */
   strcpy (return_val, home);
   return_val[home_len] = '/';
@@ -185,6 +182,7 @@ history_filename (const char *filename)
   return (return_val);
 }
 
+#if defined (DEBUG)
 static char *
 history_backupfile (const char *filename)
 {
@@ -192,7 +190,6 @@ history_backupfile (const char *filename)
   char *ret, linkbuf[PATH_MAX+1];
   size_t len;
   ssize_t n;
-  struct stat fs;
 
   fn = filename;  
 #if defined (HAVE_READLINK)
@@ -212,6 +209,7 @@ history_backupfile (const char *filename)
   ret[len+1] = '\0';
   return ret;
 }
+#endif
   
 static char *
 history_tempfile (const char *filename)
@@ -220,7 +218,6 @@ history_tempfile (const char *filename)
   char *ret, linkbuf[PATH_MAX+1];
   size_t len;
   ssize_t n;
-  struct stat fs;
   int pid;
 
   fn = filename;  
@@ -272,6 +269,7 @@ read_history_range (const char *filename, int from, int to)
   register char *line_start, *line_end, *p;
   char *input, *buffer, *bufend, *last_ts;
   int file, current_line, chars_read, has_timestamps, reset_comment_char;
+  int skipblanks, default_skipblanks;
   struct stat finfo;
   size_t file_size;
 #if defined (EFBIG)
@@ -286,7 +284,10 @@ read_history_range (const char *filename, int from, int to)
 
   buffer = last_ts = (char *)NULL;
   input = history_filename (filename);
-  file = input ? open (input, O_RDONLY|O_BINARY, 0666) : -1;
+  if (input == 0)
+    return 0;
+  errno = 0;
+  file = open (input, O_RDONLY|O_BINARY, 0666);
 
   if ((file < 0) || (fstat (file, &finfo) == -1))
     goto error_and_exit;
@@ -300,21 +301,24 @@ read_history_range (const char *filename, int from, int to)
 #endif
       goto error_and_exit;
     }
-
-  file_size = (size_t)finfo.st_size;
-
-  /* check for overflow on very large files */
-  if (file_size != finfo.st_size || file_size + 1 < file_size)
+  else
     {
-      errno = overflow_errno;
-      goto error_and_exit;
-    }
+      /* regular file */
+      file_size = (size_t)finfo.st_size;
 
-  if (file_size == 0)
-    {
-      free (input);
-      close (file);
-      return 0;	/* don't waste time if we don't have to */
+      /* check for overflow on very large files */
+      if (file_size != finfo.st_size || file_size + 1 < file_size)
+	{
+	  errno = overflow_errno;
+	  goto error_and_exit;
+	}
+
+      if (file_size == 0)
+	{
+	  xfree (input);
+	  close (file);
+	  return 0;	/* don't waste time if we don't have to */
+	}
     }
 
 #ifdef HISTORY_USE_MMAP
@@ -379,6 +383,9 @@ read_history_range (const char *filename, int from, int to)
   has_timestamps = HIST_TIMESTAMP_START (buffer);
   history_multiline_entries += has_timestamps && history_write_timestamps;
 
+  /* default is to skip blank lines unless history entries are multiline */
+  default_skipblanks = history_multiline_entries == 0;
+
   /* Skip lines until we are at FROM. */
   if (has_timestamps)
     last_ts = buffer;
@@ -404,6 +411,8 @@ read_history_range (const char *filename, int from, int to)
 	  }
       }
 
+  skipblanks = default_skipblanks;
+
   /* If there are lines left to gobble, then gobble them now. */
   for (line_end = line_start; line_end < bufend; line_end++)
     if (*line_end == '\n')
@@ -414,10 +423,16 @@ read_history_range (const char *filename, int from, int to)
 	else
 	  *line_end = '\0';
 
-	if (*line_start)
+	if (*line_start || skipblanks == 0)
 	  {
 	    if (HIST_TIMESTAMP_START(line_start) == 0)
 	      {
+		/* If we have multiline entries (default_skipblanks == 0), we
+		   don't want to skip blanks here, since we turned that on at
+		   the last timestamp line. Consider doing this even if
+		   default_skipblanks == 1 in order not to lose blank lines in
+		   commands. */
+		skipblanks = default_skipblanks;
 	      	if (last_ts == NULL && history_length > 0 && history_multiline_entries)
 		  _hs_append_history_line (history_length - 1, line_start);
 		else
@@ -432,6 +447,9 @@ read_history_range (const char *filename, int from, int to)
 	      {
 		last_ts = line_start;
 		current_line--;
+		/* Even if we're not skipping blank lines by default, we want
+		   to skip leading blank lines after a timestamp. */
+		skipblanks = 1;
 	      }
 	  }
 
@@ -469,6 +487,7 @@ history_rename (const char *old, const char *new)
 #endif
 }
 
+#if defined (DEBUG)
 /* Save FILENAME to BACK, handling case where FILENAME is a symlink
    (e.g., ~/.bash_history -> .histfiles/.bash_history.$HOSTNAME) */
 static int
@@ -487,6 +506,7 @@ histfile_backup (const char *filename, const char *back)
 #endif
   return (history_rename (filename, back));
 }
+#endif
 
 /* Restore ORIG from BACKUP handling case where ORIG is a symlink
    (e.g., ~/.bash_history -> .histfiles/.bash_history.$HOSTNAME) */
@@ -528,10 +548,14 @@ history_truncate_file (const char *fname, int lines)
 
   buffer = (char *)NULL;
   filename = history_filename (fname);
+  if (filename == 0)
+    return 0;
   tempname = 0;
-  file = filename ? open (filename, O_RDONLY|O_BINARY, 0666) : -1;
+  file = open (filename, O_RDONLY|O_BINARY, 0666);
   rv = exists = 0;
 
+  orig_lines = lines;
+
   /* Don't try to truncate non-regular files. */
   if (file == -1 || fstat (file, &finfo) == -1)
     {
@@ -580,6 +604,15 @@ history_truncate_file (const char *fname, int lines)
       goto truncate_exit;
     }
 
+  /* Don't bother with any of this if we're truncating to zero length. */
+  if (lines == 0)
+    {
+      close (file);
+      buffer[chars_read = 0] = '\0';
+      bp = buffer;
+      goto truncate_write;
+    }
+
   chars_read = read (file, buffer, file_size);
   close (file);
 
@@ -588,31 +621,38 @@ history_truncate_file (const char *fname, int lines)
       rv = (chars_read < 0) ? errno : 0;
       goto truncate_exit;
     }
+  buffer[chars_read] = '\0';	/* for the initial check of bp1[1] */
 
-  orig_lines = lines;
   /* Count backwards from the end of buffer until we have passed
      LINES lines.  bp1 is set funny initially.  But since bp[1] can't
      be a comment character (since it's off the end) and *bp can't be
-     both a newline and the history comment character, it should be OK. */
-  for (bp1 = bp = buffer + chars_read - 1; lines && bp > buffer; bp--)
+     both a newline and the history comment character, it should be OK.
+     If we are writing history timestamps, we need to add one to LINES
+     because we decrement it one extra time the first time through the loop
+     and we need the final timestamp line. */
+  lines += history_write_timestamps;
+  for (bp1 = bp = buffer + chars_read - 1; lines > 0 && bp > buffer; bp--)
     {
       if (*bp == '\n' && HIST_TIMESTAMP_START(bp1) == 0)
 	lines--;
       bp1 = bp;
     }
 
-  /* If this is the first line, then the file contains exactly the
+  /* This is the right line, so move to its start. If we're writing history
+     timestamps, we want to go back until *bp == '\n' and bp1 starts a
+     history timestamp. If we're not, just move back to *bp == '\n'.
+     If this is the first line, then the file contains exactly the
      number of lines we want to truncate to, so we don't need to do
-     anything.  It's the first line if we don't find a newline between
-     the current value of i and 0.  Otherwise, write from the start of
-     this line until the end of the buffer. */
+     anything, and we'll end up with bp == buffer.
+     Otherwise, write from the start of this line until the end of the
+     buffer. */
   for ( ; bp > buffer; bp--)
     {
-      if (*bp == '\n' && HIST_TIMESTAMP_START(bp1) == 0)
-        {
+      if (*bp == '\n' && (history_write_timestamps == 0 || HIST_TIMESTAMP_START(bp1)))
+	{
 	  bp++;
 	  break;
-        }
+	}
       bp1 = bp;
     }
 
@@ -626,17 +666,25 @@ history_truncate_file (const char *fname, int lines)
       goto truncate_exit;
     }
 
+truncate_write:
   tempname = history_tempfile (filename);
 
+  rv = 0;
   if ((file = open (tempname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0600)) != -1)
     {
       if (write (file, bp, chars_read - (bp - buffer)) < 0)
-	rv = errno;
+	{
+	  rv = errno;
+	  close (file);
+	}
 
-      if (fstat (file, &nfinfo) < 0 && rv == 0)
-	rv = errno;
+      if (rv == 0 && fstat (file, &nfinfo) < 0)
+	{
+	  rv = errno;
+	  close (file);
+	}
 
-      if (close (file) < 0 && rv == 0)
+      if (rv == 0 && close (file) < 0)
 	rv = errno;
     }
   else
@@ -673,6 +721,38 @@ history_truncate_file (const char *fname, int lines)
   return rv;
 }
 
+/* Use stdio to write the history file after mmap or malloc fails, on the
+   assumption that the stdio library can allocate the smaller buffers it uses. */
+static int
+history_write_slow (int fd, HIST_ENTRY **the_history, int nelements, int overwrite)
+{
+  FILE *fp;
+  int i, e;
+
+  fp = fdopen (fd, overwrite ? "w" : "a");
+  if (fp == 0)
+    return -1;
+
+  for (i = history_length - nelements; i < history_length; i++)
+    {
+      if (history_write_timestamps && the_history[i]->timestamp && the_history[i]->timestamp[0])
+	fprintf (fp, "%s\n", the_history[i]->timestamp);
+      if (fprintf (fp, "%s\n", the_history[i]->line) < 0)
+	goto slow_write_error;
+    }
+  if (fflush (fp) < 0)
+    {
+slow_write_error:
+      e = errno;
+      fclose (fp);
+      errno = e;
+      return -1;
+    }
+  if (fclose (fp) < 0)
+    return -1;
+  return 0;
+}
+
 /* Workhorse function for writing history.  Writes the last NELEMENT entries
    from the history list to FILENAME.  OVERWRITE is non-zero if you
    wish to replace FILENAME with the entries. */
@@ -682,12 +762,16 @@ history_do_write (const char *filename, int nelements, int overwrite)
   register int i;
   char *output, *tempname, *histname;
   int file, mode, rv, exists;
-  struct stat finfo, nfinfo;
+  struct stat finfo;
 #ifdef HISTORY_USE_MMAP
-  size_t cursize;
+  size_t cursize, newsize;
+  off_t offset;
 
   history_lines_written_to_file = 0;
 
+  if (nelements < 0)
+    return (0);
+
   mode = overwrite ? O_RDWR|O_CREAT|O_TRUNC|O_BINARY : O_RDWR|O_APPEND|O_BINARY;
 #else
   mode = overwrite ? O_WRONLY|O_CREAT|O_TRUNC|O_BINARY : O_WRONLY|O_APPEND|O_BINARY;
@@ -720,8 +804,8 @@ history_do_write (const char *filename, int nelements, int overwrite)
      Suggested by Peter Ho (peter@robosts.oxford.ac.uk). */
   {
     HIST_ENTRY **the_history;	/* local */
-    register int j;
-    int buffer_size;
+    size_t j;
+    size_t buffer_size;
     char *buffer;
 
     the_history = history_list ();
@@ -737,10 +821,16 @@ history_do_write (const char *filename, int nelements, int overwrite)
 #ifdef HISTORY_USE_MMAP
     if (ftruncate (file, buffer_size+cursize) == -1)
       goto mmap_error;
-    buffer = (char *)mmap (0, buffer_size, PROT_READ|PROT_WRITE, MAP_WFLAGS, file, cursize);
+    /* for portability, ensure that we round cursize to a multiple of the
+       page size. */
+    offset = cursize & ~(getpagesize () - 1);
+    newsize = buffer_size + cursize - offset;
+    buffer = (char *)mmap (0, newsize, PROT_READ|PROT_WRITE, MAP_WFLAGS, file, offset);
     if ((void *)buffer == MAP_FAILED)
       {
 mmap_error:
+	if ((rv = history_write_slow (file, the_history, nelements, overwrite)) == 0)
+	  goto write_success;
 	rv = errno;
 	close (file);
 	if (tempname)
@@ -749,10 +839,13 @@ mmap_error:
 	FREE (tempname);
 	return rv;
       }
+    j = cursize - offset;
 #else    
     buffer = (char *)malloc (buffer_size);
     if (buffer == 0)
       {
+	if ((rv = history_write_slow (file, the_history, nelements, overwrite)) == 0)
+	  goto write_success;
       	rv = errno;
 	close (file);
 	if (tempname)
@@ -761,9 +854,10 @@ mmap_error:
 	FREE (tempname);
 	return rv;
       }
+    j = 0;
 #endif
 
-    for (j = 0, i = history_length - nelements; i < history_length; i++)
+    for (i = history_length - nelements; i < history_length; i++)
       {
 	if (history_write_timestamps && the_history[i]->timestamp && the_history[i]->timestamp[0])
 	  {
@@ -777,7 +871,10 @@ mmap_error:
       }
 
 #ifdef HISTORY_USE_MMAP
-    if (msync (buffer, buffer_size, MS_ASYNC) != 0 || munmap (buffer, buffer_size) != 0)
+    /* make sure we unmap the pages even if the sync fails */
+    if (msync (buffer, newsize, MS_ASYNC) != 0)
+      rv = errno;
+    if (munmap (buffer, newsize) != 0 && rv == 0)
       rv = errno;
 #else
     if (write (file, buffer, buffer_size) < 0)
@@ -791,6 +888,7 @@ mmap_error:
   if (close (file) < 0 && rv == 0)
     rv = errno;
 
+write_success:
   if (rv == 0 && histname && tempname)
     rv = histfile_restore (tempname, histname);
 
